/*
  MM5 file reader for C
  written by Shigenori OTSUKA 
  from 2006.12.20         
  last midified 2008.9.3
*/

#include "mm5_io.h"

/*initialize MM5_STRUCT*/
void init_mm5struct(struct MM5_STRUCT *mm5){
  long *i;
  mm5->file = NULL;
  mm5->currenttime = 0;
  mm5->currentvar = 0;
  for(i=mm5->pos_of_time;i<mm5->pos_of_time+MAX_XTIME;i++)
    *i=-1;
  memset(mm5->name_of_vars, 0, VARNAME_LEN*MAX_VARS*MAX_XTIME);
  mm5->xtimes_flag = 0;
  mm5->ntime_flag = 0;
  mm5->bh.done = 0;
}

/*swap long*/
inline int32_t swapint32(int32_t *data){
  return(( ((*data) & 0xFF000000) >> 24)
	 +(((*data) & 0x00FF0000) >>  8)
	 +(((*data) & 0x0000FF00) <<  8)
	 +(((*data) & 0x000000FF) << 24));
}

/*swap float*/
inline float swapfloat(float *data){
  char *tmpdata;
  char ctmp[4];
  float tmp;
  if (sizeof(float)==4){
    tmpdata = (char*)data;
    ctmp[0] = tmpdata[3];
    ctmp[1] = tmpdata[2];
    ctmp[2] = tmpdata[1];
    ctmp[3] = tmpdata[0];
    tmp = *(float*)&ctmp;
    return tmp;
  }
  else{
    fprintf(stderr,"internal error: sizeof(float)!=4\n");
  }
}

void mm5_rewind(struct MM5_STRUCT *mm5){
  rewind(mm5->file);
  mm5->currenttime = 0;
  mm5->currentvar = 0;
}

/*read iheader*/
int32_t iheader(struct MM5_STRUCT *mm5){
  int32_t ihd, size;
  //fseek(mm5->file,4,SEEK_CUR);
#ifdef MM5DEBUG 
  //fprintf(stderr,"feof=%d\n",feof(mm5->file));
#endif
  if(fread(&size,4,1,mm5->file)!=1){
#ifdef MM5DEBUG    
    //fprintf(stderr,"EOF, feof=%d\n",feof(mm5->file));
#endif
    return -1;
  }
#ifdef LITTLE_ENDIAN
  size = swapint32(&size);
#endif
  if(size!=4){
    fprintf(stderr,"ERROR!!! attempt to read iheader with wrong position\n");
    return -1;
  }
  fread(&ihd,4,1,mm5->file);
  fread(&size,4,1,mm5->file);
  //fseek(mm5->file,4,SEEK_CUR);
#ifdef LITTLE_ENDIAN
  return swapint32(&ihd);
#else
  return ihd;
#endif
}

/*read bigheader*/
int read_bigheader(struct MM5_STRUCT *mm5){
  int32_t *i;
  float *j;

  if(mm5->bh.done) return 0;

  mm5_rewind(mm5);
  if(iheader(mm5)!=0) return -1;

  fseek(mm5->file,4,SEEK_CUR);

  /*read BHI*/
  fread(mm5->bh.bhi,4,50*20,mm5->file);
  /*read BHR*/
  fread(mm5->bh.bhr,4,20*20,mm5->file);
  /*read descriptions*/
  fread(mm5->bh.bhic,1,50*20*80,mm5->file);
  fread(mm5->bh.bhrc,1,20*20*80,mm5->file);

  fseek(mm5->file,4,SEEK_CUR);

  /*convert endian*/
#ifdef LITTLE_ENDIAN
  for(i=mm5->bh.bhi;i<mm5->bh.bhi+50*20;i++)
    *i = swapint32(i);
  for(j=mm5->bh.bhr;j<mm5->bh.bhr+20*20;j++)
    *j = swapfloat(j);
#endif
  mm5->bh.done = 1;
  return 0;
}

/*read subheader*/
void read_subheader(struct MM5_STRUCT *mm5,struct MM5_SUBHEADER *sh){
  int32_t *i;

  fseek(mm5->file,4,SEEK_CUR);

  fread(&(sh->ndim),4,1,mm5->file);
  fread(sh->start_index,4,4,mm5->file);
  fread(sh->end_index,4,4,mm5->file);
  fread(&(sh->xtime),4,1,mm5->file);

  fread(sh->staggering,1,4,mm5->file);
  fread(sh->ordering,1,4,mm5->file);
  fread(sh->current_time,1,24,mm5->file);
  fread(sh->name,1,9,mm5->file);
  fread(sh->unit,1,25,mm5->file);
  fread(sh->description,1,46,mm5->file);

  fseek(mm5->file,4,SEEK_CUR);

  /*convert endian*/
#ifdef LITTLE_ENDIAN
  sh->ndim = swapint32(&(sh->ndim));
  for(i=sh->start_index;i<sh->start_index+4;i++)
    *i = swapint32(i);
  for(i=sh->end_index;i<sh->end_index+4;i++)
    *i = swapint32(i);
  sh->xtime = swapfloat(&(sh->xtime));
#endif
}

/*read field data of current filepoint*/
void read_field(struct MM5_STRUCT *mm5,float *data,size_t total){
  float *i;
#ifdef MM5DEBUG
  fprintf(stderr,"total=%d\n",total);
#endif

  fseek(mm5->file,4,SEEK_CUR);

  fread(data,sizeof(float),total,mm5->file);
#ifdef LITTLE_ENDIAN
  for (i=data;i<data+total;i++)
    *i = swapfloat(i);
#endif
  fseek(mm5->file,4,SEEK_CUR);
}


/*skip record*/
void skip_record(mm5)
     struct MM5_STRUCT *mm5;
{
  int32_t size;
  fread(&size,4,1,mm5->file);
#ifdef LITTLE_ENDIAN
  size = swapint32(&size);
#endif
  size += 4;
  fseek(mm5->file,size,SEEK_CUR);
}

/*read var name in sub header*/
void read_varname(struct MM5_STRUCT *mm5,char *name){
  int32_t pos;
  char buffer[10];
  char *i;
  memset(buffer, 0, 10);
  pos = ftell(mm5->file);
  fseek(mm5->file,4+4+4*4+4*4+4+4+4+24,SEEK_CUR);
  fread(buffer,1,9,mm5->file);
  /*cleanup spaces*/
  for(i=buffer+8; i>=buffer; i--){
    if((*i != 0x20) && (*i != 0)){
      *(i+1) = 0;
      break;
    }
  }
  strcpy(name,buffer);
  /*jump to the original position*/
  fseek(mm5->file,pos,SEEK_SET);
}

/*seek top of a time "t"*/
int seek_time(struct MM5_STRUCT *mm5, int targettime){
  /*check cache data*/
  if (mm5->pos_of_time[targettime]>-1){
    fseek(mm5->file,mm5->pos_of_time[targettime],SEEK_SET);
    mm5->currenttime = targettime;
    mm5->currentvar = 0;
    return 0;
  }
  /*seek for "time"*/
  else{
    if(targettime <= (mm5->currenttime)) mm5_rewind(mm5);
    do{
      switch(iheader(mm5)){
      case 0:
	skip_record(mm5); 
	if((mm5->currenttime)==targettime){
	  mm5->pos_of_time[targettime] = ftell(mm5->file);
	  return 0;
	}
	break;
      case 1:
	skip_record(mm5);
	skip_record(mm5); 
	(mm5->currentvar)++;
	break;
      case 2:
	fgetc(mm5->file);
	if(feof(mm5->file)){
	  fprintf(stderr,"couldn't find time %d\n",targettime);
	  mm5_rewind(mm5);
	  return -1;
	}
	fseek(mm5->file,-1,SEEK_CUR);
	(mm5->currenttime)++;
	mm5->currentvar = 0;
	if((mm5->currenttime)==targettime){
	  mm5->pos_of_time[targettime] = ftell(mm5->file);
	  return 0;
	}
	if (mm5->currenttime == MAX_XTIME){
	  fprintf(stderr,"number of xtime is too big\n");
	  return -2;
	}
	break;
      default:
	/*error*/
	mm5_rewind(mm5);
	fprintf(stderr,"ERROR!!! iheader was wrong\n");
	return -3;
	break;
      }
    }while(!feof(mm5->file));
    if(!mm5->ntime_flag){
      mm5->ntime = mm5->currenttime + 1;
      mm5->ntime_flag = 1;
    }
    fprintf(stderr,"couldn't find time %d\n",targettime);
    mm5_rewind(mm5);
    return -1;
  }
}

/*seek top of a field "var" within current xtime*/
int seek_var(mm5,varname)
     struct MM5_STRUCT *mm5;
     char *varname;
{
  char name[VARNAME_LEN];
  char *currentname;

  //@pos_var={} if !defined?(@pos_var)
  //@pos_var[varname]=[] if @pos_var[varname]==nil
  //if @pos_var[varname][t]!=nil
  //@mm5file.seek(@pos_var[varname][t],whence=IO::SEEK_SET)
  //@sh = read_subheader
  //return @sh.dup
  //else
#ifdef MM5DEBUG
  fprintf(stderr,"currenttime=%d,currentvar=%d\n",mm5->currenttime,mm5->currentvar);
#endif
  do{
    switch(iheader(mm5)){
    case 0:
      skip_record(mm5); 
      if((mm5->pos_of_time[mm5->currenttime])<0){
	mm5->pos_of_time[mm5->currenttime] = ftell(mm5->file);
      }
      break;
    case 1:
      /*check cache data*/
      currentname = mm5->name_of_vars 
	+ (mm5->currentvar+(mm5->currenttime)*MAX_VARS)*VARNAME_LEN;
      if(*(currentname)!=0){
#ifdef MM5DEBUG
	fprintf(stderr,"cache %s(%d)\n",currentname,mm5->currenttime);
	//if(compare_varname(varname,currentname)) fprintf(stderr,"hit\n");
	if(!strcmp(varname,currentname)) fprintf(stderr,"hit\n");
#endif
	if(!strcmp(varname,currentname)) return 0;
      }
      /*read varname*/
      else{
	read_varname(mm5,name);
	//mm5->pos_of_vars[(mm5->currentvar)+(mm5->currenttime)*MAX_VARS] 
	//= ftell(mm5->file);
	if(mm5->currenttime<MAX_VARS)
	  strncpy(currentname,name,VARNAME_LEN-1);
#ifdef MM5DEBUG
 	fprintf(stderr,"read %s(%d)\n",name,mm5->currenttime);
	if(mm5->currenttime<MAX_VARS)
	  fprintf(stderr,"save var name of currenttime=%d,currentvar=%d\n",mm5->currenttime,mm5->currentvar);
	//if(compare_varname(varname,name)) fprintf(stderr,"hit\n");
	if(!strcmp(varname,currentname)) fprintf(stderr,"hit\n");
#endif
	if(!strcmp(varname,currentname)) return 0;
      }
      skip_record(mm5);/*subheader*/
      skip_record(mm5);/*field*/
      (mm5->currentvar)++;
      break;
    case 2:
      (mm5->currenttime)++;
      mm5->currentvar = 0;
      return -2;
      break;
    default:
      /*error*/
      mm5_rewind(mm5);
      fprintf(stderr,"ERROR!!! iheader was wrong\n");
      return -3;
      break;
    }
  }while(!feof(mm5->file));
  if(!mm5->ntime_flag){
    mm5->ntime = mm5->currenttime + 1;
    mm5->ntime_flag = 1;
  }
  fprintf(stderr,"couldn't find var %s\n",varname);
  mm5_rewind(mm5);
  return -1;
}

int count_ntime(struct MM5_STRUCT *mm5){
  /*check cache data*/
  if(mm5->ntime_flag) return mm5->ntime;
  for(mm5->currenttime=0;mm5->currenttime<MAX_XTIME;mm5->currenttime++){
    if(mm5->pos_of_time[mm5->currenttime]<0){
      if(mm5->currenttime>0){
	(mm5->currenttime)--;
	fseek(mm5->file,mm5->pos_of_time[mm5->currenttime],SEEK_SET);
      }
      else{
	mm5_rewind(mm5);
      }
      break;
    }
  }
  mm5->currentvar = 0;
  do{
    switch(iheader(mm5)){
    case 0:
      skip_record(mm5); 
      if((mm5->pos_of_time[mm5->currenttime])<0){
	mm5->pos_of_time[mm5->currenttime] = ftell(mm5->file);
      }
      break;
    case 1:
      skip_record(mm5);
      skip_record(mm5); 
      (mm5->currentvar)++;
      break;
    case 2:
      fgetc(mm5->file);
      if(feof(mm5->file)){
	mm5->ntime = mm5->currenttime + 1;
	mm5->ntime_flag = 1;
	mm5_rewind(mm5);
	return 0;	
      }
      fseek(mm5->file,-1,SEEK_CUR);
      (mm5->currenttime)++;
      mm5->currentvar = 0;
      if((mm5->pos_of_time[mm5->currenttime])<0){
	mm5->pos_of_time[mm5->currenttime] = ftell(mm5->file);
      }
      if (mm5->currenttime == MAX_XTIME){
	fprintf(stderr,"number of xtime is too big\n");
	return -2;
      }
      break;
    default:
      /*error*/
      //fprintf(stderr,"file pointer=%d\n",ftell(mm5->file));
      //fseek(mm5->file,0,SEEK_END);
      //fprintf(stderr,"file pointer end=%d\n",ftell(mm5->file));
      mm5->ntime_flag = 1;
      mm5->ntime = mm5->currenttime;
      mm5_rewind(mm5);
      fprintf(stderr,"ERROR!!! iheader was wrong\n");
      return -3;
      break;
    }
  }while(!feof(mm5->file));
  mm5->ntime_flag = 1;
  mm5->ntime = mm5->currenttime + 1;
  mm5_rewind(mm5);
  return 0;
}

/*read all xtimes*/
int read_xtimes(struct MM5_STRUCT *mm5){
  struct MM5_SUBHEADER sh;
  int done;

  mm5_rewind(mm5);
  do{
    done = 0;
    switch(iheader(mm5)){
    case 0:
      skip_record(mm5); 
      if((mm5->pos_of_time[mm5->currenttime])<0){
	mm5->pos_of_time[mm5->currenttime] = ftell(mm5->file);
      }
      break;
    case 1:
      if(!done){
	read_subheader(mm5,&sh);
	mm5->xtimes[mm5->currenttime] = sh.xtime;
	done = 1;
      }
      else{
	skip_record(mm5);
      }
      skip_record(mm5); 
      (mm5->currentvar)++;
      break;
    case 2:
      fgetc(mm5->file);
      if(feof(mm5->file)){
	if(!mm5->ntime_flag){
	  mm5->ntime = mm5->currenttime + 1;
	  mm5->ntime_flag = 1;
	}
	mm5_rewind(mm5);
	return 0;	
      }
      fseek(mm5->file,-1,SEEK_CUR);
      (mm5->currenttime)++;
      mm5->currentvar = 0;
      if((mm5->pos_of_time[mm5->currenttime])<0){
	mm5->pos_of_time[mm5->currenttime] = ftell(mm5->file);
      }
      if (mm5->ntime_flag)
	if(mm5->ntime == mm5->currenttime){
	  mm5_rewind(mm5);
	  return 0;	
	}
      if (mm5->currenttime == MAX_XTIME){
	fprintf(stderr,"number of xtime is too big\n");
	return -2;
      }
      break;
    default: 
      /*error*/
      //fprintf(stderr,"file pointer=%d\n",ftell(mm5->file));
      //fseek(mm5->file,0,SEEK_END);
      //fprintf(stderr,"file pointer end=%d\n",ftell(mm5->file));
      mm5->ntime = mm5->currenttime;
      mm5->ntime_flag = 1;
      mm5_rewind(mm5);
      fprintf(stderr,"ERROR!!! iheader was wrong\n");
      return -3;
      break;
    }
  }while(!feof(mm5->file));
  if(!mm5->ntime_flag){
    mm5->ntime = mm5->currenttime + 1;
    mm5->ntime_flag = 1;
  }
  mm5_rewind(mm5);
  return 0;
}

/*return all variable names*/
int mm5_read_all_names_of_vars(struct MM5_STRUCT *mm5)
{
  char name[10];
  char *currentname;

  mm5_rewind(mm5);
  do{
    switch(iheader(mm5)){
    case 0:
      //printf("bh\n");
      skip_record(mm5); 
      if((mm5->pos_of_time[mm5->currenttime])<0){
	mm5->pos_of_time[mm5->currenttime] = ftell(mm5->file);
      }
      break;
    case 1:
      //printf("field\n");
      /*check cache data*/
      currentname = mm5->name_of_vars 
	+ (mm5->currentvar+(mm5->currenttime)*MAX_VARS)*VARNAME_LEN;
      if(*(currentname)==0){
	/*read varname*/
	read_varname(mm5,name);
	if(mm5->currenttime<MAX_VARS)
	  strncpy(currentname,name,VARNAME_LEN-1);
      }
      skip_record(mm5);/*subheader*/
      skip_record(mm5);/*field*/
      (mm5->currentvar)++;
      break;
    case 2:
      //printf("end of time\n");
      //printf("eof=%d\n",feof(mm5->file));
      (mm5->currenttime)++;
      mm5->currentvar = 0;
      break;
    default:
      //printf("eof=%d\n",feof(mm5->file));
      if(feof(mm5->file)) break;
      /*error*/
      mm5_rewind(mm5);
      fprintf(stderr,"ERROR!!! iheader was wrong\n");
      return -3;
      break;
    }
  }while(!feof(mm5->file));
  if(!mm5->ntime_flag){
    mm5->ntime = mm5->currenttime + 1;
    mm5->ntime_flag = 1;
  }
  mm5_rewind(mm5);
  return 0;
}


/*read field data of current filepoint*/
void write_field(struct MM5_STRUCT *mm5,float *data,size_t total){
  float *i;
  int32_t tmp_size;
  size_t result;

  tmp_size = total * sizeof(float);
#ifdef LITTLE_ENDIAN
  tmp_size = swapint32(&tmp_size);
#endif

#ifdef MM5DEBUG
  fprintf(stderr,"total=%d\n",total);
#endif

  result = fwrite(&tmp_size,sizeof(int32_t),1,mm5->file);
#ifdef MM5DEBUG
  fprintf(stderr,"num of written data: %d x %d\n",sizeof(int32_t),result);
#endif

#ifdef LITTLE_ENDIAN
  for (i=data;i<data+total;i++)
    *i = swapfloat(i);
  result = fwrite(data,sizeof(float),total,mm5->file);
  for (i=data;i<data+total;i++)
    *i = swapfloat(i);
#else
  result = fwrite(data,sizeof(float),total,mm5->file);
#endif
#ifdef MM5DEBUG
  fprintf(stderr,"num of written data: %d x %d\n",sizeof(float),result);
#endif

  result = fwrite(&tmp_size,sizeof(int32_t),1,mm5->file);
#ifdef MM5DEBUG
  fprintf(stderr,"num of written data: %d x %d\n",sizeof(int32_t),result);
#endif

  fdatasync(mm5->file);
  return;
}

