/*
  MM5 file reader for C
  written by Shigenori OTSUKA 
  from 2006.12.20         
  last midified 2008.9.3

  usage example:

  struct MM5_STRUCT mm5;
  int_4 targettime=0;
  float *data;
  //
  init_mm5struct(mm5);
  mm5.file = fopen("filename","r");
  seek_time(mm5,targettime);
  seek_var(mm5,"varname");
  read_subheader(mm5);
  total = 1;
  for(i=0;i<4;i++){
    total *= ((mm5.sh.end_index + i)*(mm5.sh.start_index + i));
  }
  read_field(mm5,data,total);

*/
#include <stdio.h>
#include <string.h>
#include "ruby.h"

#define MAX_XTIME 1000
#define MAX_VARS 100
#define VARNAME_LEN 10

struct MM5_BIGHEADER{
  int32_t bhi[50*20];
  float bhr[20*20];
  char bhic[50*20*80];
  char bhrc[20*20*80];
  int done;
};

struct MM5_SUBHEADER{
  int32_t ndim;
  int32_t start_index[4];
  int32_t end_index[4];
  float xtime;
  char staggering[5];   /*  4-> 5 FOR NULL TERMINATION */
  char ordering[5];     /*  4-> 5 FOR NULL TERMINATION */
  char current_time[25];/* 24->25 FOR NULL TERMINATION */
  char name[9];         /* SPACE FOR NULL ALREADY INCLUDED */
  char unit[26];        /* 25->26 FOR NULL TERMINATION */
  char description[47]; /* 46->47 FOR NULL TERMINATION */
};

struct MM5_STRUCT {
  FILE *file;
  struct MM5_BIGHEADER bh;
  int currenttime;
  int currentvar;
  long pos_of_time[MAX_XTIME];
  //int32_t pos_of_vars[MAX_VARS,MAX_XTIME];
  char name_of_vars[VARNAME_LEN*MAX_VARS*MAX_XTIME];
  int xtimes_flag;
  float xtimes[MAX_XTIME];
  int ntime_flag;
  int ntime;
};

void init_mm5struct(struct MM5_STRUCT *mm5);

int32_t swapint32(int32_t *data);
float swapfloat(float *data);

void mm5_rewind(struct MM5_STRUCT *mm5);
int32_t iheader(struct MM5_STRUCT *mm5);
int read_bigheader(struct MM5_STRUCT *mm5);
void read_subheader(struct MM5_STRUCT *mm5,struct MM5_SUBHEADER *sh);
void read_field(struct MM5_STRUCT *mm5,float *data,size_t total);
void skip_record(struct MM5_STRUCT *mm5);
void read_varname(struct MM5_STRUCT *mm5,char *name);
int seek_time(struct MM5_STRUCT *mm5,int targettime);
int seek_var(struct MM5_STRUCT *mm5,char *varname);
int count_ntime(struct MM5_STRUCT *mm5);
int read_xtimes(struct MM5_STRUCT *mm5);
int mm5_read_all_names_of_vars(struct MM5_STRUCT *mm5);
void write_field(struct MM5_STRUCT *mm5,float *data,size_t total);
