/*
  MM5 file reader for C
  written by Shigenori OTSUKA 
  from 2006.12.20         
  last midified 2008.9.3
*/

#include "mm5_io.h"
#include "mm5_io_wrap.h"

VALUE cMM5IO, cNArray;

static VALUE rb_mm5_make_mm5()
{
  struct MM5_STRUCT *mm5 = ALLOC(struct MM5_STRUCT);
  return Data_Wrap_Struct(cMM5IO, NULL,rb_mm5_free_mm5, mm5);
}

static void rb_mm5_free_mm5(struct MM5_STRUCT *mm5){
  if (mm5->file) 
    if(fclose(mm5->file)) printf("error while closing file %p!!\n",mm5->file);
  xfree(mm5);
}

static VALUE rb_mm5_initialize(VALUE self){
  struct MM5_STRUCT *mm5;
  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
  /*initialize values*/
  init_mm5struct(mm5);
  return Qnil;
}

static VALUE rb_mm5_singleton_open(int argc, VALUE argv[], VALUE self){
  VALUE mm5;
  mm5 = rb_funcall(self, rb_intern("new"), 0);
  rb_mm5_open(argc, argv, mm5);
  return mm5;
}

static VALUE rb_mm5_new_ruby16()
{
  struct MM5_STRUCT *mm5 = ALLOC(struct MM5_STRUCT);
  init_mm5struct(mm5);
  return Data_Wrap_Struct(cMM5IO, NULL,rb_mm5_free_mm5, mm5);
}

static VALUE rb_mm5_open(int argc, VALUE argv[], VALUE self){
  char *name;
  struct MM5_STRUCT *mm5;
  int i;
  char *mode;

  if(argc>2)
    rb_raise(rb_eRuntimeError,"too much args; %d for 2", argc);
  if(argc==0)
    rb_raise(rb_eRuntimeError,"at least 1 arg is needed");

  Data_Get_Struct(self, struct MM5_STRUCT, mm5);

  for(i=0; i<2; i++)
    if(rb_class_of(argv[0])!=rb_cString){
      rb_raise(rb_eRuntimeError,"arg %d is not a String", i);
      return Qfalse;
    }

  name = StringValuePtr(argv[0]);

  if(argc==1)
    mode = "rb";
  else{
    rb_funcall(argv[1], rb_intern("strip!"), 0);
    mode = StringValuePtr(argv[1]);
    if(strcmp(mode,"rb") || strcmp(mode,"r+b"))
      ;
    else if(strcmp(mode,"r"))
      mode = "rb";
    else if(strcmp(mode,"r+"))
      mode = "r+b";
    else{
      rb_raise(rb_eRuntimeError,"file mode is limited to r, r+, rb, and r+b");
    }
  }

  if(mm5->file)
    mm5->file = freopen(name, mode ,mm5->file);
  else
    mm5->file = fopen(name, mode);

  if (mm5->file==NULL){
    rb_raise(rb_eRuntimeError,"couldn't open file: %s\n",name);
    return Qfalse;
  }
  return self;
}

static VALUE rb_mm5_close(VALUE self){
  struct MM5_STRUCT *mm5;
  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
  if(mm5->file)
    if(fclose(mm5->file)==EOF) fprintf(stderr, "EOF!!\n");
  init_mm5struct(mm5);
  return Qnil;
}

static VALUE rb_mm5_get_shape(VALUE self, VALUE varname, VALUE time){
  char *name;
  int targettime;
  struct MM5_STRUCT *mm5;
  struct MM5_SUBHEADER sh = {0,{0},{0},0,{0},{0},{0},{0},{0},{0}};
  VALUE shape;
  int i;

  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
  if(mm5->file==NULL)
    rb_raise(rb_eRuntimeError,"file not opened\n");

  name = StringValuePtr(varname);
  targettime = (int)NUM2INT(time);

#ifdef MM5DEBUG
  fprintf(stderr, "seek_time: %d\n",targettime);
#endif
  if(seek_time(mm5,targettime)) 
    rb_raise(rb_eRuntimeError,"couldn't find t=%d\n",targettime);
#ifdef MM5DEBUG
  fprintf(stderr, "seek_var: %s\n",name);
#endif
  if(seek_var(mm5,name))
    rb_raise(rb_eRuntimeError,"couldn't find var %s\n",name);
#ifdef MM5DEBUG
  fprintf(stderr, "read_subheader\n");
#endif
  read_subheader(mm5,&sh);

  /* create array */
  shape = rb_ary_new();
  if((strncmp(sh.staggering,"C",1)==0) && (sh.ndim > 1)){
    rb_ary_push(shape, INT2NUM(sh.end_index[1] - sh.start_index[1] + 1 - 1));
    rb_ary_push(shape, INT2NUM(sh.end_index[0] - sh.start_index[0] + 1 - 1));
    if(sh.ndim>2)
      for(i=2; i<sh.ndim; i++)
	rb_ary_push(shape, INT2NUM(sh.end_index[i] - sh.start_index[i] + 1));
  }
  else if((strncmp(sh.staggering,"D",1)==0) && (sh.ndim > 1)){
    rb_ary_push(shape, INT2NUM(sh.end_index[1] - sh.start_index[1] + 1));
    rb_ary_push(shape, INT2NUM(sh.end_index[0] - sh.start_index[0] + 1));
    if(sh.ndim>2)
      for(i=2; i<sh.ndim; i++)
	rb_ary_push(shape, INT2NUM(sh.end_index[i] - sh.start_index[i] + 1));
  }  
  else{
    for(i=0; i<sh.ndim; i++)
      rb_ary_push(shape, INT2NUM(sh.end_index[i] - sh.start_index[i] + 1));
  }

  return shape;
}

static VALUE rb_mm5_staggering(VALUE self, VALUE varname, VALUE time){
  char *name;
  int targettime;
  struct MM5_STRUCT *mm5;
  struct MM5_SUBHEADER sh = {0,{0},{0},0,{0},{0},{0},{0},{0},{0}};
  VALUE shape;
  int i;

  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
  if(mm5->file==NULL)
    rb_raise(rb_eRuntimeError,"file not opened\n");

  name = StringValuePtr(varname);
  targettime = (int)NUM2INT(time);
  if(seek_time(mm5,targettime)) 
    rb_raise(rb_eRuntimeError,"couldn't find t=%d\n",targettime);
  if(seek_var(mm5,name))
    rb_raise(rb_eRuntimeError,"couldn't find var %s\n",name);
  read_subheader(mm5,&sh);

  return rb_str_new2(sh.staggering);
}

static VALUE rb_mm5_ordering(VALUE self, VALUE varname, VALUE time){
  char *name;
  int targettime;
  struct MM5_STRUCT *mm5;
  struct MM5_SUBHEADER sh = {0,{0},{0},0,{0},{0},{0},{0},{0},{0}};
  VALUE shape;
  int i;

  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
  if(mm5->file==NULL)
    rb_raise(rb_eRuntimeError,"file not opened\n");

  name = StringValuePtr(varname);
  targettime = (int)NUM2INT(time);
  if(seek_time(mm5,targettime)) 
    rb_raise(rb_eRuntimeError,"couldn't find t=%d\n",targettime);
  if(seek_var(mm5,name))
    rb_raise(rb_eRuntimeError,"couldn't find var %s\n",name);
  read_subheader(mm5,&sh);

  return rb_str_new2(sh.ordering);
}

static VALUE rb_mm5_read(VALUE self, VALUE varname, VALUE time){
  char *name;
  int targettime;
  struct MM5_STRUCT *mm5;
  struct MM5_SUBHEADER sh = {0,{0},{0},0,{0},{0},{0},{0},{0},{0}};
  int shape[4],shape0,shape1,shape2;
  size_t total;
  VALUE vdata;
  struct NARRAY *ndata;
  float *pdata,*pdatatmp;
  int i,j,k;
  float *pfloat,*pfloats; 

  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
  if(mm5->file==NULL)
    rb_raise(rb_eRuntimeError,"file not opened\n");

  name = StringValuePtr(varname);
  targettime = (int)NUM2INT(time);

#ifdef MM5DEBUG
  fprintf(stderr, "seek_time: %d\n",targettime);
#endif
  if(seek_time(mm5,targettime)) 
    rb_raise(rb_eRuntimeError,"couldn't find t=%d\n",targettime);
#ifdef MM5DEBUG
  fprintf(stderr, "seek_var: %s\n",name);
#endif
  if(seek_var(mm5,name))
    rb_raise(rb_eRuntimeError,"couldn't find var %s\n",name);
#ifdef MM5DEBUG
  fprintf(stderr, "read_subheader\n");
#endif
  read_subheader(mm5,&sh);

  /*calculate size of data*/
  total = 1;
  for(i=0;i<4;i++){
    shape[i] = sh.end_index[i] - sh.start_index[i] + 1;
    total *= shape[i];
  }
#ifdef MM5DEBUG
  fprintf(stderr, "shape=%d,%d,%d,%d\n",shape[0],shape[1],shape[2],shape[3]);  
  fprintf(stderr, "total=%d\n",total);  
#endif

  /*allocate working area*/
  pdatatmp = ALLOC_N(float, total);

  /*read field*/
#ifdef MM5DEBUG
  fprintf(stderr, "read_field\n");
#endif
  read_field(mm5,pdatatmp,total);

  /*convert data*/
#ifdef MM5DEBUG
  fprintf(stderr, "convert shape...\n");
#endif
  if ((strncmp(sh.staggering,"C",1)==0) && (sh.ndim > 1)){
    /*create a NArray object*/
    shape0 = shape[1] - 1;
    shape1 = shape[0] - 1;
    shape2 = shape[2];
    shape[0] = shape0;
    shape[1] = shape1;
#ifdef MM5DEBUG
  fprintf(stderr, "shape=%d,%d,%d,%d\n",shape[0],shape[1],shape[2],shape[3]);  
#endif
    vdata = na_make_object(NA_SFLOAT,sh.ndim,shape,cNArray);
    GetNArray(vdata,ndata);
    pdata = (float*)ndata->ptr;
#ifdef MM5DEBUG
    fprintf(stderr, "  rank=%d\n",(int)ndata->rank);
    fprintf(stderr, "  total=%d\n",(int)ndata->total);
    fprintf(stderr, "  type=%d\n",(int)ndata->type);
    fprintf(stderr, "  shape=%d,%d,%d\n",(int)ndata->shape[0],(int)ndata->shape[1],(int)ndata->shape[2]);
#endif

    pfloat = pdatatmp;
    pfloats = pdata;
    switch(sh.ndim){
    case 2:
#ifdef MM5DEBUG
      fprintf(stderr, "CROSS 2D...");
#endif
      for(j=0;j<shape0;j++){
	pfloats = pdata + j;
	for(i=0;i<shape1;i++){
	  *pfloats = (*pfloat);
	  pfloat++;
	  pfloats += shape0;
	}
	pfloat++;
      }
      break;
    case 3:
#ifdef MM5DEBUG
      fprintf(stderr, "CROSS 3D...");
#endif
      for(k=0;k<shape2;k++){
	for(j=0;j<shape0;j++){
	  pfloats = pdata + j + k*shape0*shape1;
	  for(i=0;i<shape1;i++){
#ifdef MM5DEBUG
	    //fprintf(stderr, "%p,%p\n",pfloats,pdata + 4 * ((int)ndata->total - 1));
	    //fprintf(stderr, "i,j,k=%d,%d,%d\n",i,j,k);
#endif
	    *pfloats = (*pfloat);
	    pfloat++;
	    pfloats += shape0;
	  }
	  pfloat++;
	}
	pfloat += (shape1+1);
      }
      break;
    }
#ifdef MM5DEBUG
    fprintf(stderr, "done\n");
#endif
    /*free work area*/
    free(pdatatmp);
    return vdata;
  }
  else if((strncmp(sh.staggering,"D",1)==0) && (sh.ndim > 1)){
    /*create a NArray object*/
    shape0 = shape[1];
    shape1 = shape[0];
    shape2 = shape[2];
    shape[0] = shape0;
    shape[1] = shape1;
    vdata = na_make_object(NA_SFLOAT,sh.ndim,shape,cNArray);
    GetNArray(vdata,ndata);
    pdata = (float*)ndata->ptr;

    pfloat = pdatatmp;
    pfloats = pdata;
    switch(sh.ndim){
    case 2:
#ifdef MM5DEBUG
      fprintf(stderr, "DOT 2D...");
#endif
      for(j=0;j<shape0;j++){
	pfloats = pdata + j;
	for(i=0;i<shape1;i++){
	  *pfloats = (*pfloat);
	  pfloat++;
	  pfloats += shape0;
	}
      }
      break;
    case 3:
#ifdef MM5DEBUG
      fprintf(stderr, "DOT 3D...");
#endif
      for(k=0;k<shape2;k++){
	for(j=0;j<shape0;j++){
	  pfloats = pdata + j + k*shape0*shape1;
	  for(i=0;i<shape1;i++){
	    *pfloats = (*pfloat);
	    pfloat++;
	    pfloats += shape0;
	  }
	}
      }
      break;
    }
#ifdef MM5DEBUG
    fprintf(stderr, "done\n");
#endif
    /*free work area*/
    free(pdatatmp);
    return vdata;

  }
  else{
#ifdef MM5DEBUG
    fprintf(stderr, "else...");
#endif
    /*create a NArray object*/
    //fprintf(stderr, "ndim=%d...",sh.ndim);
    //fprintf(stderr, "shape[0]=%d...",shape[0]);
    //fprintf(stderr,"make na...");  
    vdata = na_make_object(NA_SFLOAT,sh.ndim,shape,cNArray);
    GetNArray(vdata,ndata);
    pdata = (float*)ndata->ptr;
    memcpy(pdata,pdatatmp,sizeof(float)*total);
#ifdef MM5DEBUG
    fprintf(stderr, "done\n");
#endif
    /*free work area*/
    free(pdatatmp);
    return vdata;
  }
}

static VALUE rb_mm5_bhi(VALUE self,VALUE vi,VALUE vj){
  struct MM5_STRUCT *mm5;
  int i,j;

  i = NUM2INT(vi);
  j = NUM2INT(vj);
  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
#ifdef MM5DEBUG
  fprintf(stderr, "bhi\n");
#endif
  read_bigheader(mm5);
  return(INT2FIX(mm5->bh.bhi[i+j*50]));
}

static VALUE rb_mm5_bhr(VALUE self,VALUE vi,VALUE vj){
  struct MM5_STRUCT *mm5;
  int i,j;

  i = NUM2INT(vi);
  j = NUM2INT(vj);
  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
#ifdef MM5DEBUG
  fprintf(stderr, "bhr\n");
#endif
  read_bigheader(mm5);
  return rb_float_new(mm5->bh.bhr[i+j*20]);
}

static VALUE rb_mm5_ntime(VALUE self){
  struct MM5_STRUCT *mm5;
  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
#ifdef MM5DEBUG
  fprintf(stderr, "count_ntime\n");
#endif
  count_ntime(mm5);
  return INT2FIX(mm5->ntime);
}

static VALUE rb_mm5_xtimes(VALUE self){
  struct MM5_STRUCT *mm5;
  VALUE xtimes;
  int i;

  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
  if(!mm5->ntime_flag) count_ntime(mm5);
#ifdef MM5DEBUG
  fprintf(stderr, "read_xtimes\n");
#endif
  read_xtimes(mm5);
  //xtimes = rb_ary_new2((int32_t)(mm5->ntime));
  xtimes = rb_ary_new();
  for(i=0;i<mm5->ntime;i++){
    xtimes = rb_ary_push(xtimes,rb_float_new(mm5->xtimes[i]));
  }
  return xtimes;
}

static VALUE rb_mm5_var_names(VALUE self){
  struct MM5_STRUCT *mm5;
  int i,j;
  char *name;
  VALUE ary;

  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
  mm5_read_all_names_of_vars(mm5);

  /* convert C strings to Ruby Array */
  ary = rb_ary_new();
  for(i=0; i<(mm5->ntime); i++){
    for(j=0; j<MAX_VARS; j++){
      name = mm5->name_of_vars 
	+(j+i*MAX_VARS)*VARNAME_LEN;
      if(*name!=0)
	rb_ary_push(ary, rb_str_new2(name));
    }
  }
  rb_funcall(ary, rb_intern("uniq!"), 0);
  return ary;
}

static VALUE rb_mm5_overwrite(VALUE self, VALUE varname, VALUE time, VALUE data){
  char *name;
  int targettime;
  struct MM5_STRUCT *mm5;
  struct MM5_SUBHEADER sh = {0,{0},{0},0,{0},{0},{0},{0},{0},{0}};
  int shape[4],shape0,shape1,shape2;
  size_t total;
  //VALUE vdata;
  struct NARRAY *ndata;
  float *pdata,*pdatatmp;
  int i,j,k;
  float *pfloat,*pfloats; 

  Data_Get_Struct(self, struct MM5_STRUCT, mm5);
  if(mm5->file==NULL)
    rb_raise(rb_eRuntimeError,"file not opened\n");

  name = StringValuePtr(varname);
  targettime = (int)NUM2INT(time);
  data = na_cast_object(data,NA_SFLOAT);
  GetNArray(data,ndata);
  pdata = (float*)ndata->ptr;

#ifdef MM5DEBUG
  fprintf(stderr, "seek_time: %d\n",targettime);
#endif
  if(seek_time(mm5,targettime)) 
    rb_raise(rb_eRuntimeError,"couldn't find t=%d\n",targettime);
#ifdef MM5DEBUG
  fprintf(stderr, "seek_var: %s\n",name);
#endif
  if(seek_var(mm5,name))
    rb_raise(rb_eRuntimeError,"couldn't find var %s\n",name);
#ifdef MM5DEBUG
  fprintf(stderr, "read_subheader\n");
#endif
  read_subheader(mm5,&sh);

  /*calculate size of data*/
  total = 1;
  for(i=0;i<4;i++){
    shape[i] = sh.end_index[i] - sh.start_index[i] + 1;
    total *= shape[i];
  }
#ifdef MM5DEBUG
  fprintf(stderr, "shape=%d,%d,%d,%d\n",shape[0],shape[1],shape[2],shape[3]);  
  fprintf(stderr, "total=%d\n",total);  
#endif

  /*allocate working area*/
  pdatatmp = ALLOC_N(float, total);

  /*convert data*/
#ifdef MM5DEBUG
  fprintf(stderr, "convert shape...\n");
#endif
  if ((strncmp(sh.staggering,"C",1)==0) && (sh.ndim > 1)){
    shape0 = shape[1] - 1;
    shape1 = shape[0] - 1;
    shape2 = shape[2];
    pfloat = pdatatmp; // data to be written on file
    pfloats = pdata;   // data on ruby
    switch(sh.ndim){
    case 2:
#ifdef MM5DEBUG
      fprintf(stderr, "CROSS 2D...");
#endif
      for(j=0;j<shape0;j++){
	pfloats = pdata + j;
	for(i=0;i<shape1;i++){
	  //*pfloats = (*pfloat);
	  *pfloat = (*pfloats);
	  pfloat++;
	  pfloats += shape0;
	}
	pfloat++;
      }
      break;
    case 3:
#ifdef MM5DEBUG
      fprintf(stderr, "CROSS 3D...");
#endif
      for(k=0;k<shape2;k++){
	for(j=0;j<shape0;j++){
	  pfloats = pdata + j + k*shape0*shape1;
	  for(i=0;i<shape1;i++){
#ifdef MM5DEBUG
	    //fprintf(stderr, "%p,%p\n",pfloats,pdata + 4 * ((int)ndata->total - 1));
	    //fprintf(stderr, "i,j,k=%d,%d,%d\n",i,j,k);
#endif
	    //*pfloats = (*pfloat);
	    *pfloat = (*pfloats);
	    pfloat++;
	    pfloats += shape0;
	  }
	  pfloat++;
	}
	pfloat += (shape1+1);
      }
      break;
    }
#ifdef MM5DEBUG
    fprintf(stderr, "done\n");
#endif
  }
  else if((strncmp(sh.staggering,"D",1)==0) && (sh.ndim > 1)){
    shape0 = shape[1];
    shape1 = shape[0];
    shape2 = shape[2];
    pfloat = pdatatmp; // data to be written on file
    pfloats = pdata;   // data on ruby
    switch(sh.ndim){
    case 2:
#ifdef MM5DEBUG
      fprintf(stderr, "DOT 2D...");
#endif
      for(j=0;j<shape0;j++){
	pfloats = pdata + j;
	for(i=0;i<shape1;i++){
	  //*pfloats = (*pfloat);
	  *pfloat = (*pfloats);
	  pfloat++;
	  pfloats += shape0;
	}
      }
      break;
    case 3:
#ifdef MM5DEBUG
      fprintf(stderr, "DOT 3D...");
#endif
      for(k=0;k<shape2;k++){
	for(j=0;j<shape0;j++){
	  pfloats = pdata + j + k*shape0*shape1;
	  for(i=0;i<shape1;i++){
	    //*pfloats = (*pfloat);
	    *pfloat = (*pfloats);
	    pfloat++;
	    pfloats += shape0;
	  }
	}
      }
      break;
    }
#ifdef MM5DEBUG
    fprintf(stderr, "done\n");
#endif
  }
  else{
#ifdef MM5DEBUG
    fprintf(stderr, "else...");
#endif
    //memcpy(pdata,pdatatmp,sizeof(float)*total);
    memcpy(pdatatmp,pdata,sizeof(float)*total);
#ifdef MM5DEBUG
    fprintf(stderr, "done\n");
#endif
  }

  /*write field*/
#ifdef MM5DEBUG
  fprintf(stderr, "write_field\n");
#endif
  write_field(mm5,pdatatmp,total);
  free(pdatatmp);
  return Qnil;
}

/*register functions to ruby interpreter*/
void Init_mm5_io_raw()
{
  rb_require("narray");
  cNArray = rb_const_get( rb_cObject, rb_intern("NArray") );

  cMM5IO = rb_define_class("MM5IO",rb_cObject);
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_define_alloc_func(cMM5IO, rb_mm5_make_mm5);
  rb_define_private_method(cMM5IO,"initialize",rb_mm5_initialize,0);
#else
  rb_define_singleton_method(cMM5IO,"new",rb_mm5_new_ruby16,0);
#endif
  rb_define_method(cMM5IO,"open",rb_mm5_open,-1);
  rb_define_singleton_method(cMM5IO,"open",rb_mm5_singleton_open,-1);
  rb_define_method(cMM5IO,"close",rb_mm5_close,0);
  /*methods for read*/
  rb_define_method(cMM5IO,"read",rb_mm5_read,2);
  rb_define_method(cMM5IO,"get_shape",rb_mm5_get_shape,2);
  rb_define_method(cMM5IO,"staggering",rb_mm5_staggering,2);
  rb_define_method(cMM5IO,"ordering",rb_mm5_ordering,2);
  rb_define_method(cMM5IO,"bhi",rb_mm5_bhi,2);
  rb_define_method(cMM5IO,"bhr",rb_mm5_bhr,2);
  rb_define_method(cMM5IO,"ntime",rb_mm5_ntime,0);
  rb_define_method(cMM5IO,"xtimes",rb_mm5_xtimes,0);
  rb_define_method(cMM5IO,"var_names",rb_mm5_var_names,0);
  /*methods for write*/
  rb_define_method(cMM5IO,"overwrite",rb_mm5_overwrite,3);
}
